(*
   Copyright 2012 by Mark Weyer
*)

type polynomial = bool array

let zero = ([| |] : polynomial)

let crc32 = [|
  true;  true;  true;  false; true;  true;  false; true;
  true;  false; true;  true;  true;  false; false; false;
  true;  false; false; false; false; false; true;  true;
  false; false; true;  false; false; false; false; false;
  true;
  |]

class type crc_run = object
  method read_bit : bool -> unit
  method read_byte_little_endian : int -> unit
  method remainder : polynomial
end

class constant_polynomial_run = object
  method read_bit (_:bool) = ()
  method read_byte_little_endian (_:int) = ()
  method remainder = zero
end

let unoptimized_run polynomial remainder =
    (* The degree must be nonzero and the highest coefficient must be 1. *)
  let n = Array.length polynomial - 1  in
  object (self)

  method read_bit bit =
    let sub = remainder.(n-1) <> bit  in
    Array.blit remainder 0 remainder 1 (n-1);
    remainder.(0) <- false;
    if sub
    then for i=0 to n-1 do remainder.(i) <- remainder.(i)<>polynomial.(i) done

  method read_byte_little_endian byte =
    for i=0 to 7 do
      self#read_bit ((byte lsr i) land 1 = 1)
    done

  method remainder = remainder

end



type spec = {
  polynomial : polynomial;
  initial_remainder : polynomial;
  }

class type crc = object
  method start_run : crc_run
end

class constant_polynomial = object
  val run = new constant_polynomial_run
  method start_run = run
end

class unoptimized spec =
  let irl = Array.length spec.initial_remainder  in
  let initial = Array.init (Array.length spec.polynomial - 1)
    (fun i -> if i<irl
      then spec.initial_remainder.(i)
      else false)  in
  object
  method start_run = unoptimized_run spec.polynomial (Array.copy initial)
end



let rec crc spec =
  let n = Array.length spec.polynomial - 1  in
  if n<0
  then raise Division_by_zero
  else if not spec.polynomial.(n)
    then crc {spec with polynomial=Array.sub spec.polynomial 0 n}
    else if n=0
      then new constant_polynomial
      else new unoptimized spec

