(*
   Copyright 2012 by Mark Weyer
*)


let encode_int16 string pos int =
  string.[pos] <- char_of_int (int/256);
  string.[pos+1] <- char_of_int (int mod 256)

let encode_float string pos float = encode_int16 string pos
  (int_of_float (floor ((max 0.0 (min 1.0 float))*.65535.0 +. 0.5)))

let encode_int32 string pos int32 =
  for i=0 to 3 do
    string.[pos+3-i] <- char_of_int (Int32.to_int
      (Int32.logand 255l (Int32.shift_right_logical int32 (8*i))));
  done

let int32buffer = "...."

let output_int32 file int32 =
  encode_int32 int32buffer 0 int32;
  output_string file int32buffer


let crc_spec = {
  Crc.polynomial = Crc.crc32;
  Crc.initial_remainder = Array.make 32 true;
  }

let crc = Crc.crc crc_spec

let crc data =
  let run = crc#start_run  in
  String.iter (fun char -> run#read_byte_little_endian (int_of_char char)) data;
  Int32.logxor 0xffffffffl
    (Array.fold_left
      (fun r bit ->
        let r' = Int32.shift_left r 1  in
        if bit then Int32.succ r' else r')
      0l
      run#remainder)


let output_chunk file name data =
  let len = String.length data  in
  output_int32 file (Int32.of_int len);
  output_string file name;
  output_string file data;
  output_int32 file (crc (name^data))


let ihdr_chunk = "wwwwhhhh\016\006\000\000\000"

let output_ihdr file width height =
  encode_int32 ihdr_chunk 0 (Int32.of_int width);
  encode_int32 ihdr_chunk 4 (Int32.of_int height);
  output_chunk file "IHDR" ihdr_chunk


let line_data line =
  let len = Array.length line  in
  let data = String.create (8*len+1)  in
  data.[0] <- '\000';
  Array.iteri (fun i farbe ->
    encode_float data (8*i+1) (Farbe.nur_rot farbe);
    encode_float data (8*i+3) (Farbe.nur_gruen farbe);
    encode_float data (8*i+5) (Farbe.nur_blau farbe);
    encode_float data (8*i+7) (1.0 -. Farbe.nur_durchsichtig farbe))
    line;
  data


let output_png name pixelbild =
  let file = open_out name  in
  let width,height,colours = pixelbild  in
  output_string file "\137PNG\013\010\026\010";
  output_ihdr file width height;
  let zlib_header,zlib = Zlib.encode ()  in
  output_chunk file "IDAT" zlib_header;
  Array.iteri
    (fun i line ->
      let data = line_data line  in
      output_chunk file "IDAT" (zlib#more_data (i=height-1) data))
    colours;
  output_chunk file "IDAT" zlib#trailer;
  output_chunk file "IEND" "";
  close_out file

